# -*- Mode: Python; tab-width: 4 -*-
#       Id: asynchat.py,v 2.26 2000/09/07 22:29:26 rushing Exp
#       Author: Sam Rushing <rushing@nightmare.com>

# ======================================================================
# Copyright 1996 by Sam Rushing
#
#                         All Rights Reserved
#
# Permission to use, copy, modify, and distribute this software and
# its documentation for any purpose and without fee is hereby
# granted, provided that the above copyright notice appear in all
# copies and that both that copyright notice and this permission
# notice appear in supporting documentation, and that the name of Sam
# Rushing not be used in advertising or publicity pertaining to
# distribution of the software without specific, written prior
# permission.
#
# SAM RUSHING DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
# INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN
# NO EVENT SHALL SAM RUSHING BE LIABLE FOR ANY SPECIAL, INDIRECT OR
# CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
# NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
# ======================================================================

r"""A class supporting chat-style (command/response) protocols.

This class adds support for 'chat' style protocols - where one side
sends a 'command', and the other sends a response (examples would be
the common internet protocols - smtp, nntp, ftp, etc..).

The handle_read() method looks at the input stream for the current
'terminator', calling self.found_terminator() on its receipt.

for example:
Say you build an async nntp client using this class.  At the start of
the connection, you'll have called self.set_terminator(b'\r\n'), in
order to process the single-line greeting.  Just before issuing a LIST
command you'll set it to b'\r\n.\r\n'.  The output of the LIST command
will be accumulated (using your own 'collect_incoming_data' method) up
to the terminator, and then control will be returned to you - by
calling your self.found_terminator() method.

"""

import sys
import socket
import asyncore
from collections import deque

class async_chat (asyncore.dispatcher):

    """This is an abstract class.  You must derive from this class,
    and add the `found_terminator` method. Children may also override
    `collect_incoming_data` to override the default buffering of
    incoming data.

    The set_terminator(term) or set_terminator_str(term, encoding)
    method almost always should be called before entering
    asyncore.loop.

    In addition to overriding methods, children may override the
    variables `ac_in_buffer_size` and `ac_out_buffer_size`, which
    provide hints regarding the number of bytes that may be stored in
    the incoming and outgoing buffers, respectively.  Changing these
    will not affect functionality, but may alter performance.  The
    default value of each is 4096.

    """

    # these are overridable defaults

    ac_in_buffer_size       = 4096
    ac_out_buffer_size      = 4096

    def __init__ (self, conn=None):
        """Constructor, should be called by child's __init__

        Optional arguments:
        conn -- An already initialized socket.

        conn, if provided and not None, will be used as the socket for
        this channel. If conn is provided, create_socket and its
        related methods need not be called.

        """

        # for terminator matching
        self.__in_buffer = b''
        # we use a list here rather than cStringIO for a few reasons...
        # del lst[:] is faster than sio.truncate(0)
        # lst = [] is faster than sio.truncate(0)
        # cStringIO will be gaining unicode support in py3k, which
        # will negatively affect the performance of bytes compared to
        # a ''.join() equivalent
        self.__incoming = []

        # we toss the use of the "simple producer" and replace it with
        # a pure deque, which the original fifo was a wrapping of
        self.__producer_fifo = deque()

        asyncore.dispatcher.__init__ (self, conn)

    def collect_incoming_data(self, data):
        """Store received data for later processing

        Arguments:
        data -- A bytes object representing some received data.

        collect_incoming_data may be overridden in a child class.  The
        default implementation simply records incoming data, making it
        accessible via the `get_data` method; `found_terminator` is
        the proper place to do processing in most cases.

        """
        self.__incoming.append(data)

    def get_data(self):
        """Retrieve the data received so far. Returns a bytes object.

        This method is tied to the default collect_incoming_data
        implementation. It retrieves and empties the buffer that is
        filled by the default collect_incoming_data.

        """
        d = b''.join(self.__incoming)
        del self.__incoming[:]
        return d

    def found_terminator(self):
        """Process a complete section of data received from the socket

        found_terminator must be overridden in a child class for
        async_chat to be useful.  It is called when the terminator is
        found in the received data stream, after calling
        `collect_incoming_data` with all bytes prior to the terminator
        in the stream. The terminator is defined by the most recent
        call to `set_terminator` or `set_terminator_str`.  If
        `collect_incoming_data` has not been overridden, get_data()
        will return the data received up to the terminator.


        """
        raise NotImplementedError("found_terminator must be implemented in subclass")

    def set_terminator (self, term):
        """Set the input delimiter.

        Arguments:
        term -- Can be a bytes object, an integer, or None.

        If term is None, async_chat will never call
        found_terminator. This is occasionally useful, if
        `collect_incoming_data` performs all necessary handling of the
        data.

        If term is an int object, async_chat will read that many bytes
        from the data stream, and then call `found_terminator`.  This
        is useful for reading protocol sections of known length.

        If term is neither an int nor None, it will be passed to the
        bytes object constructor.  async_chat will look for that
        sequence of bytes in the incoming data stream, and call
        `found_terminator` when it is discovered.  This is useful for
        reading protocol sections of varying or unknown length.

        set_terminator may be called as many times as you like.  This
        is quite often useful -- for example in HTTP you might read
        the headers by doing set_terminator(b'\n\n') and then extract
        the content length from the headers and do
        set_terminator(content_length) to read the page data.

        """

        if term is None:
            self.terminator = None
        elif isinstance(term, int):
            self.terminator = term
        else:
            self.terminator = bytes(term)

    def set_terminator_str(self, term, encoding):
        """Set the input delimiter to an encoded string

        Arguments:
        term -- A str object.
        encoding -- The name of a codec.

        This method is like `set_terminator`, except that it sets the
        terminator to a sequence of bytes representing term in the
        given encoding.

        It is very important to be aware of byte ordering when using
        this function. Some encodings produce endian-sentive results.

        """
        self.set_terminator(term.encode(encoding))

    def get_terminator (self):
        """Return the current terminator"""
        return self.terminator

    # grab some more data from the socket,
    # throw it to the collector method,
    # check for the terminator,
    # if found, transition to the next state.

    def handle_read (self):
        """asyncore function. Overriding children must invoke."""

        try:
            data = self.recv(self.ac_in_buffer_size)
        except socket.error:
            self.handle_error()
            return

        data = bytes(data)
        self.__in_buffer = self.__in_buffer + data

        # Continue to search for self.terminator in self.__in_buffer,
        # while calling self.collect_incoming_data.  The while loop
        # is necessary because we might read several data+terminator
        # combos with a single recv(4096).

        while self.__in_buffer:
            lb = len(self.__in_buffer)
            terminator = self.get_terminator()
            if not terminator:
                # no terminator, collect it all
                self.collect_incoming_data (self.__in_buffer)
                self.__in_buffer = b''
            elif isinstance(terminator, int):
                # numeric terminator
                n = terminator
                if lb < n:
                    self.collect_incoming_data (self.__in_buffer)
                    self.__in_buffer = b''
                    self.terminator = self.terminator - lb
                else:
                    self.collect_incoming_data (self.__in_buffer[:n])
                    self.__in_buffer = self.__in_buffer[n:]
                    self.terminator = 0
                    self.found_terminator()
            else:
                # 3 cases:
                # 1) end of buffer matches terminator exactly:
                #    collect data, transition
                # 2) end of buffer matches some prefix:
                #    collect data to the prefix
                # 3) end of buffer does not match any prefix:
                #    collect data
                terminator_len = len(terminator)
                index = self.__in_buffer.find(terminator)
                if index != -1:
                    # we found the terminator
                    if index > 0:
                        # don't bother reporting the empty string (source of subtle bugs)
                        self.collect_incoming_data (self.__in_buffer[:index])
                    self.__in_buffer = self.__in_buffer[index+terminator_len:]
                    # This does the Right Thing if the terminator is changed here.
                    self.found_terminator()
                else:
                    # check for a prefix of the terminator
                    index = find_prefix_at_end (self.__in_buffer, terminator)
                    if index:
                        if index != lb:
                            # we found a prefix, collect up to the prefix
                            self.collect_incoming_data (self.__in_buffer[:-index])
                            self.__in_buffer = self.__in_buffer[-index:]
                        break
                    else:
                        # no prefix, collect it all
                        self.collect_incoming_data (self.__in_buffer)
                        self.__in_buffer = b''

    def handle_write (self):
        """asyncore function. Overriding children must invoke."""
        self.initiate_send()

    def handle_close (self):
        """asyncore function. Overriding children must invoke."""
        self.close()

    def push (self, data):
        """Place data in the outgoing queue.

        Arguments:
        data -- The expression bytes(data) must be meaningful

        Appends bytes(data) to the sequence of bytes to be sent
        through the socket.  Each time asyncore.loop discovers that
        the socket is writable, some or all of the bytes in that queue
        will be transmitted.

        """
        sabs = self.ac_out_buffer_size
        data = bytes(data)
        if len(data) > sabs:
            for i in xrange(0, len(data), sabs):
                self.__producer_fifo.append(data[i:i+sabs])
        else:
            self.__producer_fifo.append(data)
        self.initiate_send()

    def push_str(self, data, encoding):
        """Place a string in the outgoing queue.

        Arguments:
        data -- The expression data.encode(encoding) must be meaningful

        Appends bytes(data.encode(encoding)) to the sequence of bytes
        to be sent through the socket.  Each time asyncore.loop
        discovers that the socket is writable, some or all of the
        bytes in that queue will be transmitted.

        It is very important to be aware of byte ordering when using
        this function. Some encodings produce endian-sentive results.

        """
        self.push(data.encode(encoding))

    def push_iterable(self, iterable):
        """Place the iterated items in the outgoing queue

        Arguments:
        iterable -- [bytes(x) for x in iterable] must be meaningful

        Appends bytes(x) for each x in iterator to the sequence of
        bytes to be sent through the socket.  Each time asyncore.loop
        discovers that the socket is writable, some or all of the
        bytes in that queue will be transmitted.

        """
        self.__producer_fifo.append(iter(iterable))
        self.initiate_send()

    def push_with_producer (self, producer):
        """Place a producer in the outgoing queue.

        Arguments:
        producer -- The expression producer.more() must be meaningful

        Repeatedly appends bytes(producer.more()) to the sequence of
        bytes to be sent through the socket, until
        bool(producer.more()) is False.  Each time asyncore.loop
        discovers that the socket is writable, some or all of the
        bytes in that queue will be transmitted.

        """
        def adapter(prod):
            while True:
                data = prod.more()
                if data == b'':
                    raise StopIteration
                yield data

        self.push_iterable(adapter(producer))

    def readable (self):
        "predicate for inclusion in the readable for select()"
        # cannot use the old predicate, it violates the claim of the
        # set_terminator method.

        # return (len(self.__in_buffer) <= self.ac_in_buffer_size)
        return 1

    def writable (self):
        "predicate for inclusion in the writable for select()"
        return self.__producer_fifo or (not self.connected)

    def close_when_done (self):
        "automatically close this channel once the outgoing queue is empty"
        self.__producer_fifo.append(None)

    def initiate_send (self):
        """Internal function.  Do not invoke or override."""
        while self.__producer_fifo and self.connected:
            source = self.__producer_fifo.popleft()

            # Handle None entry.
            if source is None:
                self.handle_close()
                return

            # Handle empty entries.
            if not source:
                continue

            try:
                data = source.__next__()
                self.__producer_fifo.appendleft(source)
            except StopIteration:
                continue
            except AttributeError:
                data = source

            data = bytes(data)

            # Send the data.
            try:
                num_sent = self.send(data[:self.ac_out_buffer_size])
            except socket.error:
                self.handle_error()
                return

            if num_sent:
                if num_sent < len(data):
                    self.__producer_fifo.appendleft(data[num_sent:])

            # We tried to send some actual data.
            return

    def discard_buffers (self):
        """Discard all incoming and outgoing data.

        It is generally wise to avoid the use of this function, as it
        results in data loss. This is rarely useful.

        """
        # Emergencies only!
        self.__in_buffer = b''
        del self.__incoming[:]
        self.__producer_fifo.clear()

class simple_producer:

    """Produce bytes for transmission.

    This is a very basic producer; the only extra functionality it
    provides over async_chat.push(bytes) is to break the data up into
    blocks of arbitrary size, such that each call to more produces a
    block of data no larger than the requested size. Even that is not
    very useful, since setting async_chat.ac_out_buffer_size achieves
    the same thing.

    """

    def __init__ (self, data, buffer_size=512):
        """Initialize the producer with a byte sequence

        Arguments:
        data -- bytes(data) must be meaningful.

        Optional arguments:
        buffer_size -- A number, defaults to 512

        The simple_producer will pop and return blocks of `data` of
        length min(`buffer_size`, len(`data`)) until data is empty.

        """

        self.data = bytes(data)
        self.buffer_size = buffer_size

    def more (self):
        if len (self.data) > self.buffer_size:
            result = self.data[:self.buffer_size]
            self.data = self.data[self.buffer_size:]
            return result
        else:
            result = self.data
            self.data = b''
            return result

# Given 'haystack', see if any prefix of 'needle' is at its end.  This
# assumes an exact match has already been checked.  Return the number of
# characters matched.
# for example:
# f_p_a_e ("qwerty\r", "\r\n") => 1
# f_p_a_e ("qwertydkjf", "\r\n") => 0
# f_p_a_e ("qwerty\r\n", "\r\n") => <undefined>

# this could maybe be made faster with a computed regex?
# [answer: no; circa Python-2.0, Jan 2001]
# new python:   28961/s
# old python:   18307/s
# re:        12820/s
# regex:     14035/s

def find_prefix_at_end (haystack, needle):
    """Returns > 0 if the beginning of needle is at the end of haystack

    Arguments:
    haystack -- A bytes object.
    needle -- A bytes object.

    Given haystack, see if any prefix of needle is at its end.  This
    assumes an exact match has already been checked.  Return the
    number of characters matched.

    """
    l = len(needle) - 1
    while l and not haystack.endswith(needle[:l]):
        l -= 1
    return l
