# -*- Mode: Python -*-
#   Id: asyncore.py,v 2.51 2000/09/07 22:29:26 rushing Exp
#   Author: Sam Rushing <rushing@nightmare.com>

# ======================================================================
# Copyright 1996 by Sam Rushing
#
#                         All Rights Reserved
#
# Permission to use, copy, modify, and distribute this software and
# its documentation for any purpose and without fee is hereby
# granted, provided that the above copyright notice appear in all
# copies and that both that copyright notice and this permission
# notice appear in supporting documentation, and that the name of Sam
# Rushing not be used in advertising or publicity pertaining to
# distribution of the software without specific, written prior
# permission.
#
# SAM RUSHING DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
# INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN
# NO EVENT SHALL SAM RUSHING BE LIABLE FOR ANY SPECIAL, INDIRECT OR
# CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
# NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
# ======================================================================

"""Basic infrastructure for asynchronous socket service clients and servers.

There are only two ways to have a program on a single processor do "more
than one thing at a time".  Multi-threaded programming is the simplest and
most popular way to do it, but there is another very different technique,
that lets you have nearly all the advantages of multi-threading, without
actually using multiple threads. it's really only practical if your program
is largely I/O bound. If your program is CPU bound, then pre-emptive
scheduled threads are probably what you really need. Network servers are
rarely CPU-bound, however.

If your operating system supports the select() system call in its I/O
library (and nearly all do), then you can use it to juggle multiple
communication channels at once; doing other work while your I/O is taking
place in the "background."  Although this strategy can seem strange and
complex, especially at first, it is in many ways easier to understand and
control than multi-threaded programming. The module documented here solves
many of the difficult problems for you, making the task of building
sophisticated high-performance network servers and clients a snap.
"""

import select
import socket
import sys
import time

import os
from errno import (EALREADY, EINPROGRESS, EWOULDBLOCK, ECONNRESET,
                   ENOTCONN, ESHUTDOWN, EINTR, EISCONN, EISCONN,
                   EBADF, ECONNABORTED, errorcode)

try:
    socket_map
except NameError:
    socket_map = {}

def _strerror(err):
    res = os.strerror(err)
    if res == 'Unknown error':
        res = errorcode[err]
    return res

class ExitNow(Exception):
    pass

def read(obj):
    try:
        obj.handle_read_event()
    except (ExitNow, KeyboardInterrupt, SystemExit):
        raise
    except:
        obj.handle_error()

def write(obj):
    try:
        obj.handle_write_event()
    except (ExitNow, KeyboardInterrupt, SystemExit):
        raise
    except:
        obj.handle_error()

def _exception(obj):
    try:
        obj.handle_expt_event()
    except (ExitNow, KeyboardInterrupt, SystemExit):
        raise
    except:
        obj.handle_error()

def readwrite(obj, flags):
    try:
        if flags & (select.POLLIN | select.POLLPRI):
            obj.handle_read_event()
        if flags & select.POLLOUT:
            obj.handle_write_event()
        if flags & (select.POLLERR | select.POLLHUP | select.POLLNVAL):
            obj.handle_expt_event()
    except (ExitNow, KeyboardInterrupt, SystemExit):
        raise
    except:
        obj.handle_error()

def poll(timeout=0.0, map=None):
    if map is None:
        map = socket_map
    if map:
        r = []; w = []; e = []
        for fd, obj in map.items():
            is_r = obj.readable()
            is_w = obj.writable()
            if is_r:
                r.append(fd)
            if is_w:
                w.append(fd)
            if is_r or is_w:
                e.append(fd)
        if [] == r == w == e:
            time.sleep(timeout)

        try:
            r, w, e = select.select(r, w, e, timeout)
        except select.error as err:
            if err.args[0] != EINTR:
                raise
            else:
                return

        for fd in r:
            obj = map.get(fd)
            if obj is None:
                continue
            read(obj)

        for fd in w:
            obj = map.get(fd)
            if obj is None:
                continue
            write(obj)

        for fd in e:
            obj = map.get(fd)
            if obj is None:
                continue
            _exception(obj)

def poll2(timeout=0.0, map=None):
    # Use the poll() support added to the select module in Python 2.0
    if map is None:
        map = socket_map
    if timeout is not None:
        # timeout is in milliseconds
        timeout = int(timeout*1000)
    pollster = select.poll()
    if map:
        for fd, obj in map.items():
            flags = 0
            if obj.readable():
                flags |= select.POLLIN | select.POLLPRI
            if obj.writable():
                flags |= select.POLLOUT
            if flags:
                # Only check for exceptions if object was either readable
                # or writable.
                flags |= select.POLLERR | select.POLLHUP | select.POLLNVAL
                pollster.register(fd, flags)
        try:
            r = pollster.poll(timeout)
        except select.error as err:
            if err.args[0] != EINTR:
                raise
            r = []
        for fd, flags in r:
            obj = map.get(fd)
            if obj is None:
                continue
            readwrite(obj, flags)

poll3 = poll2                           # Alias for backward compatibility

def loop(timeout=30.0, use_poll=False, map=None, count=None):
    """The main event-handling loop of asyncore.

    Optional arguments:
    timeout -- Number representing seconds, default 30.0
    use_poll -- Use poll instead of select if available, default False
    map -- The map of channels to process, default is use built-in map
    count -- number of times to cycle through the loop, default infinity

    Note: Setting timeout will not cause loop to terminate after a
    fixed amount of time.  Rather, it determines how long each select
    or poll call will block.  To cause loop to terminate after a fixed
    amount of time, pass an integer count. The loop will then
    terminate after approximately count * timeout seconds, if the
    select or poll calls block.

    """

    if map is None:
        map = socket_map

    if use_poll and hasattr(select, 'poll'):
        poll_fun = poll2
    else:
        poll_fun = poll

    if count is None:
        while map:
            poll_fun(timeout, map)

    else:
        while map and count > 0:
            poll_fun(timeout, map)
            count = count - 1

class dispatcher:

    """The abstract base class that users of asyncore should inherit
    from.

    A dispatcher object handles a single socket, processing connect,
    accept, close, read and write events as defined by the child
    handle_connect, handle_accept, handle_close, handle_read and
    handle_write methods, respectively.

    The child may also define handle_expt to handle exceptional
    conditions during the communication process. Exceptional
    conditions are ill-defined, and may or may not include out of band
    data, hangup, socket error, and invalid data.

    dispatcher implements many socket methods, so children can send,
    recv, accept, bind, listen, etc within as if self were a socket.
    dispatcher also has a create_socket method for creating the
    internal socket, if necessary.

    Children should always invoke dispatcher.__init__.

    """

    debug = False
    connected = False
    accepting = False
    closing = False
    addr = None

    def __init__(self, sock=None, map=None):
        """Initialize a dispatcher.

        Children should always invoke dispatcher.__init__

        Optional arguments:
        sock -- A socket to communicate through, default None
        map -- A map to register in, default is use built-in map

        If sock is None, create_socket and either connect or
        listen/bind must be called to create and initialize the
        socket.

        If map is None, the asyncore global map is used.

        """

        if map is None:
            self._map = socket_map
        else:
            self._map = map

        self._fileno = None

        if sock:
            # Set to nonblocking just to make sure for cases where we
            # get a socket from a blocking source.
            sock.setblocking(0)
            self.set_socket(sock, map)
            self.connected = True
            # The constructor no longer requires that the socket
            # passed be connected.
            try:
                self.addr = sock.getpeername()
            except socket.error:
                if err[0] == ENOTCONN:
                    # To handle the case where we got an unconnected
                    # socket.
                    self.connected = False
                else:
                    # The socket is broken in some unknown way, alert
                    # the user and remove it from the map (to prevent
                    # polling of broken sockets).
                    self.del_channel(map)
                    raise
        else:
            self.socket = None

    def __repr__(self):
        status = [self.__class__.__module__+"."+self.__class__.__name__]
        if self.accepting and self.addr:
            status.append('listening')
        elif self.connected:
            status.append('connected')
        if self.addr is not None:
            try:
                status.append('%s:%d' % self.addr)
            except TypeError:
                status.append(repr(self.addr))
        return '<%s at %#x>' % (' '.join(status), id(self))

    def add_channel(self, map=None):
        """Adds the dispatcher to the map.

        Called automatically.

        """

        #self.log_info('adding channel %s' % self)
        if map is None:
            map = self._map
        map[self._fileno] = self

    def del_channel(self, map=None):
        """Removes the dispatcher from the map.

        Called automatically by close.

        """

        fd = self._fileno
        if map is None:
            map = self._map
        if fd in map:
            #self.log_info('closing channel %d:%s' % (fd, self))
            del map[fd]
        self._fileno = None

    def create_socket(self, family, type):
        """Creates and initializes the internal socket.

        Arguments:
        family -- socket module AF_* constant
        type -- socket module SOCK_* constant

        If an initialized socket is passed to the constructor's conn
        argument, or to the set_socket method, there is no need to
        call this method.

        """

        self.family_and_type = family, type
        sock = socket.socket(family, type)
        sock.setblocking(0)
        self.set_socket(sock)

    def set_socket(self, sock, map=None):
        """Sets the internal socket

        Arguments:
        sock -- An initialized socket

        Optional arguments:
        map -- The map to use, default is the built-in map

        """

        self.socket = sock
##        self.__dict__['socket'] = sock
        self._fileno = sock.fileno()
        self.add_channel(map)

    def set_reuse_addr(self):
        """Sets the socket.SO_REUSEADDR socket option, if possible

        This will cause the socket to attempt to re-use a server
        port. There is no guarantee of success.

        """

        # try to re-use a server port if possible
        try:
            self.socket.setsockopt(
                socket.SOL_SOCKET, socket.SO_REUSEADDR,
                self.socket.getsockopt(socket.SOL_SOCKET,
                                       socket.SO_REUSEADDR) | 1
                )
        except socket.error:
            pass

    # ==================================================
    # predicates for select()
    # these are used as filters for the lists of sockets
    # to pass to select().
    # ==================================================

    def readable(self):
        """Determine whether read events should be generated for this
        dispatcher.

        While readable returns False, read events will not be
        generated for this dispatcher, and handle_read and
        handle_accept will not be called.  handle_connect will be
        called only if triggered by a write event.

        Default is to return True.

        """
        return True

    def writable(self):
        """Determine whether write events should be generated for this
        dispatcher.

        While writable returns False, write events will not be
        generated for this dispatcher, and handle_write not be called.
        handle_connect will be called only if triggered by a read
        event.

        Default is to return True.

        """
        return True

    # ==================================================
    # socket object methods.
    # ==================================================

    def listen(self, num):
        """Behaves as the similarly-named socket method"""
        self.accepting = True
        if os.name == 'nt' and num > 5:
            num = 5
        return self.socket.listen(num)

    def bind(self, addr):
        """Behaves as the similarly-named socket method"""
        self.addr = addr
        return self.socket.bind(addr)

    def connect(self, address):
        """Behaves as the similarly-named socket method"""
        self.connected = False
        err = self.socket.connect_ex(address)
        # XXX Should interpret Winsock return values
        if err in (EINPROGRESS, EALREADY, EWOULDBLOCK):
            return
        if err in (0, EISCONN):
            self.addr = address
            self.handle_connect_event()
        else:
            raise socket.error(err, errorcode[err])

    def accept(self):
        """Behaves as the similarly-named socket method"""
        # XXX can return either an address pair or None
        try:
            conn, addr = self.socket.accept()
            return conn, addr
        except socket.error as why:
            if why[0] == EWOULDBLOCK:
                pass
            else:
                raise

    def send(self, data):
        """Behaves as the similarly-named socket method"""
        try:
            result = self.socket.send(data)
            return result
        except socket.error as why:
            if why[0] == EWOULDBLOCK:
                return 0
            elif why[0] in (ECONNRESET, ENOTCONN, ESHUTDOWN, ECONNABORTED):
                self.handle_close()
                return 0
            else:
                raise

    def recv(self, buffer_size):
        """Behaves as the similarly-named socket method"""
        try:
            data = self.socket.recv(buffer_size)
            if not data:
                # a closed connection is indicated by signaling
                # a read condition, and having recv() return 0.
                self.handle_close()
                return b''
            else:
                return data
        except socket.error as why:
            # winsock sometimes throws ENOTCONN
            if why[0] in [ECONNRESET, ENOTCONN, ESHUTDOWN, ECONNABORTED]:
                self.handle_close()
                return b''
            else:
                raise

    def close(self):
        """Behaves as the similarly-named socket method"""
        self.connected = False
        self.accepting = False
        self.del_channel()
        try:
            self.socket.close()
        except socket.error as why:
            if why[0] not in (ENOTCONN, EBADF):
                raise

    # cheap inheritance, used to pass all other attribute
    # references to the underlying socket object.
    def __getattr__(self, attr):
        return getattr(self.socket, attr)

    # log and log_info may be overridden to provide more sophisticated
    # logging and warning methods. In general, log is for 'hit' logging
    # and 'log_info' is for informational, warning and error logging.

    def log(self, message):
        sys.stderr.write('log: %s\n' % str(message))

    def log_info(self, message, type='info'):
        if __debug__ or type != 'info':
            print('%s: %s' % (type, message))

    def handle_read_event(self):
        """Internal method, do not override."""
        if self.accepting:
            # accepting sockets are never connected, they "spawn" new
            # sockets that are connected
            self.handle_accept()
        elif not self.connected:
            self.handle_connect_event()
            self.handle_read()
        else:
            self.handle_read()

    def handle_connect_event(self):
        """Internal method, do not override."""
        self.connected = True
        self.handle_connect()

    def handle_write_event(self):
        """Internal method, do not override."""
        if self.accepting:
            # Accepting sockets shouldn't get a write event.
            # We will pretend it didn't happen.
            return

        if not self.connected:
            #check for errors
            err = self.socket.getsockopt(socket.SOL_SOCKET, socket.SO_ERROR)
            if err != 0:
                raise socket.error(err, strerror(err))

            self.handle_connect_event()
        self.handle_write()

    def handle_expt_event(self):
        """Internal method, do not override."""
        self.handle_expt()

    def handle_error(self):
        """Called when an exception is raised and not
        otherwise handled.

        The default version prints a condensed traceback.

        """

        nil, t, v, tbinfo = compact_traceback()

        # sometimes a user repr method will crash.
        try:
            self_repr = repr(self)
        except:
            self_repr = '<__repr__(self) failed for object at %0x>' % id(self)

        self.log_info(
            'uncaptured python exception, closing channel %s (%s:%s %s)' % (
                self_repr,
                t,
                v,
                tbinfo
                ),
            'error'
            )
        self.close()

    def handle_expt(self):
        """Called to handle an exceptional condition event on the
        socket.

        Children may override this method to implement exceptional
        condition processing. The default behavior is to raise an
        exception.

        """
        err = self.socket.getsockopt(socket.SOL_SOCKET, socket.SO_ERROR)
        msg = _strerror(err)

        raise socket.error(err, msg)

    def handle_read(self):
        """Called to handle a read event.

        When this method is called, a call to self.recv will return
        data without blocking.  Children may override this method to
        implement data reception.

        """
        self.log_info('unhandled read event', 'warning')

    def handle_write(self):
        """Called to handle a write event.

        When this method is called, a call to self.send will transmit
        data without blocking.  Children may override this method to
        implement data transmission.

        """
        self.log_info('unhandled write event', 'warning')

    def handle_connect(self):
        """Called to handle a connect event.

        When this method is called, the socket has just been
        connected.  Calls to self.send and self.recv will succeed
        without blocking.  Children may override this method.

        """
        self.log_info('unhandled connect event', 'warning')

    def handle_accept(self):
        """Called to handle an accept event.

        When this method is called, a call to self.accept will return
        a new socket without blocking.  Children may override this
        method to implement server socket listening.

        Note: This event will not be triggered unless listen and bind
        have previously been called.  Passing a listening socket into
        the constructor or set_socket will not work.

        """
        self.log_info('unhandled accept event', 'warning')

    def handle_close(self):
        """Called to handle a close event.

        When this method is called, it is time to call self.close.
        Children may override this method to implement end-of-life
        behavior.

        """
        self.log_info('unhandled close event', 'warning')
        self.close()

# ---------------------------------------------------------------------------
# adds simple buffered output capability, useful for simple clients.
# [for more sophisticated usage use asynchat.async_chat]
# ---------------------------------------------------------------------------

class dispatcher_with_send(dispatcher):
    """A very basic buffered output implementation on top of
    dispatcher.

    The send method of this class queues data that are passed to it,
    and the handle_write method transmits data from the queue across
    the network. Thus you can simply call send with however much data
    you like, and dispatcher_with_send will see to it that all of the
    data are sent.

    The asynchat module provides a more capable implementation of
    this, along with assorted other features.

    """

    def __init__(self, sock=None, map=None):
        dispatcher.__init__(self, sock, map)
        self.out_buffer = b''

    def initiate_send(self):
        num_sent = 0
        num_sent = dispatcher.send(self, self.out_buffer[:512])
        self.out_buffer = self.out_buffer[num_sent:]

    def handle_write(self):
        self.initiate_send()

    def writable(self):
        return (not self.connected) or len(self.out_buffer)

    def send(self, data):
        assert isinstance(data, bytes)
        if self.debug:
            self.log_info('sending %s' % repr(data))
        self.out_buffer = self.out_buffer + data
        self.initiate_send()

# ---------------------------------------------------------------------------
# used for debugging.
# ---------------------------------------------------------------------------

def compact_traceback():
    t, v, tb = sys.exc_info()
    tbinfo = []
    if not tb: # Must have a traceback
        raise AssertionError("traceback does not exist")
    while tb:
        tbinfo.append((
            tb.tb_frame.f_code.co_filename,
            tb.tb_frame.f_code.co_name,
            str(tb.tb_lineno)
            ))
        tb = tb.tb_next

    # just to be safe
    del tb

    file, function, line = tbinfo[-1]
    info = ' '.join(['[%s|%s|%s]' % x for x in tbinfo])
    return (file, function, line), t, v, info

def close_all(map=None, ignore_all=False):
    if map is None:
        map = socket_map
    for x in list(map.values()):
        try:
            x.close()
        except OSError as x:
            if x[0] == EBADF:
                pass
            elif not ignore_all:
                raise
        except (ExitNow, KeyboardInterrupt, SystemExit):
            raise
        except:
            if not ignore_all:
                raise

    map.clear()

# Asynchronous File I/O:
#
# After a little research (reading man pages on various unixen, and
# digging through the linux kernel), I've determined that select()
# isn't meant for doing asynchronous file i/o.
# Heartening, though - reading linux/mm/filemap.c shows that linux
# supports asynchronous read-ahead.  So _MOST_ of the time, the data
# will be sitting in memory for us already when we go to read it.
#
# What other OS's (besides NT) support async file i/o?  [VMS?]
#
# Regardless, this is useful for pipes, and stdin/stdout...

if os.name == 'posix':
    import fcntl

    class file_wrapper:
        # Here we override just enough to make a file
        # look like a socket for the purposes of asyncore.
        # The passed fd is automatically os.dup()'d

        def __init__(self, fd):
            self.fd = os.dup(fd)

        def recv(self, *args):
            return os.read(self.fd, *args)

        def send(self, *args):
            return os.write(self.fd, *args)

        read = recv
        write = send

        def close(self):
            os.close(self.fd)

        def fileno(self):
            return self.fd

    class file_dispatcher(dispatcher):

        def __init__(self, fd, map=None):
            dispatcher.__init__(self, None, map)
            self.connected = True
            try:
                fd = fd.fileno()
            except AttributeError:
                pass
            self.set_file(fd)
            # set it to non-blocking mode
            flags = fcntl.fcntl(fd, fcntl.F_GETFL, 0)
            flags = flags | os.O_NONBLOCK
            fcntl.fcntl(fd, fcntl.F_SETFL, flags)

        def set_file(self, fd):
            self._fileno = fd
            self.socket = file_wrapper(fd)
            self.add_channel()
